(in-package :cl-utilities)

(define-condition no-extremum (error) ()
  (:report "Cannot find extremum of empty sequence")
  (:documentation "Raised when EXTREMUM is called on an empty
sequence, since there is no morally smallest element"))

(defun comparator (test &optional (key #'identity))
  "Comparison operator: auxilliary function used by EXTREMUM"
  (declare (optimize (speed 3) (safety 0) (space 0) (debug 1)))
  (lambda (a b) (if (funcall test
                             (funcall key a)
                             (funcall key b))
                    a
                    b)))

;; This optimizes the case where KEY is #'identity
(define-compiler-macro comparator (&whole whole test
					  &optional (key #'identity))
  (if (eql key #'identity)
      `(lambda (a b)
	(if (funcall ,test a b) a b))
      whole))

(defun zero-length-p (sequence)
  "Is the length of SEQUENCE equal to zero?"
  (declare (optimize (speed 3) (safety 0) (space 0) (debug 1)))
  (or (null sequence)
      (when (vectorp sequence)
	(zerop (length sequence)))))

(declaim (inline zero-length-p))

;; Checks the length of the subsequence of SEQUENCE specified by START
;; and END, and if it's 0 then a NO-EXTREMUM error is signalled. This
;; should only be used in EXTREMUM functions.
(defmacro with-check-length ((sequence start end) &body body)
  (once-only (sequence start end)
    `(if (or (zero-length-p ,sequence)
	  (>= ,start (or ,end (length ,sequence))))
      (restart-case (error 'no-extremum)
	(continue ()
	  :report "Return NIL instead"
	  nil))
      ,@body)))

;; This is an extended version which takes START and END keyword
;; arguments. Any spec-compliant use of EXTREMUM will also work with
;; this extended version.
(defun extremum (sequence predicate
		 &key (key #'identity) (start 0) end)
  "Returns the element of SEQUENCE that would appear first if the
sequence were ordered according to SORT using PREDICATE and KEY. See
http://www.cliki.net/EXTREMUM for the full
specification. Additionally, START and END specify the beginning and
ending indices of the part of the sequence we should look at."
  (with-check-length (sequence start end)
    (reduce (comparator predicate key) sequence
	    :start start :end end)))

;; This optimizes the case where KEY is #'identity
(define-compiler-macro extremum (&whole whole sequence predicate
					&key (key #'identity) (start 0) end)
  (if (eql key #'identity)
      (once-only (sequence predicate start end)
	`(with-check-length (,sequence ,start ,end)
	  (locally (declare (optimize (speed 3) (safety 0) (space 0) (debug 1)))
	    (reduce (comparator ,predicate) ,sequence
		    :start ,start :end ,end))))
      whole))

;; And, for backup, here's a strictly spec-compliant version.
#+nil
(defun extremum (sequence predicate &key (key #'identity))
  "Returns the element of SEQUENCE that would appear first if the
sequence were ordered according to SORT using PREDICATE and KEY. See
http://www.cliki.net/EXTREMUM for the full specification."
  (with-check-length (sequence 0 nil)
    (reduce (comparator predicate key) sequence)))

;; This is an "optimized" version which calls KEY less. REDUCE is
;; already so optimized that this will actually be slower unless KEY
;; is expensive. And on CLISP, of course, the regular version will be
;; much faster since built-in functions are ridiculously faster than
;; ones implemented in Lisp. Be warned, this isn't as carefully tested
;; as regular EXTREMUM and there's more that could go wrong.
(defun extremum-fastkey (sequence predicate
			 &key (key #'identity) (start 0) end)
  "EXTREMUM implemented so that it calls KEY less. This is only faster
if the KEY function is so slow that calling it less often would be a
significant improvement; ordinarily it's slower."
  (declare (optimize (speed 3) (safety 0) (space 0) (debug 1)))
  (with-check-length (sequence start end)
    (let* ((smallest (elt sequence 0))
	   (smallest-key (funcall key smallest))
	   (current-index 0)
	   (real-end (or end #.(1- most-positive-fixnum))))
      (declare (type (integer 0 #.most-positive-fixnum)
		     current-index real-end start))
      (map nil #'(lambda (x)
		   (when (<= start current-index real-end)
		     (let ((x-key (funcall key x)))
		       (when (funcall predicate
				      x-key
				      smallest-key)
			 (setf smallest x)
			 (setf smallest-key x-key))))
		   (incf current-index))
	   sequence)
      smallest)))